#include <features.h>
#include <inttypes.h>
#include <sys/socket.h>
#include <netpacket/packet.h>
#include <net/if_ether.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <unistd.h>
#include <time.h>
#include <netdb.h>
#include <sys/time.h>
#ifdef __dietlibc__
#include <write12.h>
#else
static void __write1(const char* s) {
  write(1,s,strlen(s));
}
static void __write2(const char* s) {
  write(2,s,strlen(s));
}
#endif
#include <signal.h>
#include <errno.h>

/* #define NODNS */

char arprequest[100] =
  "\xff\xff\xff\xff\xff\xff"	// 0, 6, destination ethernet address
  "\0\0\0\0\0\0"		// 6, 6, source ethernet address
  "\x08\x06"			// 12, 2, protocol type (ETH_P_ARP, 0x0806)
  "\0\x01"			// 14, 2, hardware identifier (ARPHRD_ETHER, 1)
  "\x08\x00"			// 16, 2, protocol identifier (ETH_P_IP, 0x0800)
  "\x06"			// 18, 1, hardware address length
  "\x04"			// 19, 1, protocol address length
  "\x00\x01"			// 20, 2, opcode (ARPOP_REQUEST, 1)
  "\0\0\0\0\0\0"		// 22, 6, sender hardware address
  "\0\0\0\0"			// 28, 4, sender IP address
  "\0\0\0\0\0\0"		// 32, 6, target hardware address
  "\0\0\0\0"			// 38, 4, target IP address
				// 42 end of packet
  ;

static int in_checksum(unsigned short *buf, int len) {
  register long sum = 0;
  unsigned short answer = 0;
  while (len>1) { sum += *buf++; len -= 2; }
  if (len==1) {
    *(unsigned char*)(&answer) = *(unsigned char*)buf;
    sum += answer;
  }
  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);
  answer = ~sum;
  return answer;
}

static unsigned int fmt_ulong(char *dest,unsigned long i) {
  register unsigned long len,tmp,len2;
  for (len=1, tmp=i; tmp>9; ++len) tmp/=10;
  if (dest)
    for (tmp=i, dest+=len, len2=len+1; --len2; tmp/=10)
      *--dest = (tmp%10)+'0';
  return len;
}

static unsigned int fmt_str(char *out,const char *in) {
  register char* s=out;
  register const char* t=in;
  for (;;) {
    if (!*t) break; *s=*t; ++s; ++t;
  }
  return s-out;
}

static unsigned int fmt_ip4(char *s,const char ip[4]) {
  unsigned int len;
  int i;
  len = 0;
  for (i=0; i<4; ++i) {
    register unsigned int j;
    len+=(j=fmt_ulong(s,(unsigned long) (unsigned char) ip[i]))+1;
    if (s && i<3) { s+=j; *s++='.'; }
  }
  return len-1;
}

static unsigned int scan_ulong(const char *src,unsigned long *dest) {
  register const char *tmp=src;
  register int l=0;
  register unsigned char c;
  while ((c=*tmp-'0')<10) {
    l=l*10+c;
    ++tmp;
  }
  *dest=l;
  return tmp-src;
}

static void dumpethaddr(const unsigned char* x) {
  char buf[100];
  int i,j;
  for (i=j=0; i<6; ++i) {
    j+=fmt_xlong(buf+j,x[i]);
    if (i<5) {
      buf[j]=':';
      j+=1;
    }
  }
  write(1,buf,j);
}

char* peer="127.0.0.1";
int sent,received;

main(int argc,char* argv[]) {
  int s=socket(PF_PACKET,SOCK_RAW,htons(0x806));
  char ibuf[1500];
  struct sockaddr_ll sin,sout;
  int count=-1,len,sinlen;
  socklen_t sl;
  fd_set ifd;
  struct timeval tv,start,now;
  int interface=if_nametoindex("eth0");
  int c;
  long missed;

  if (s<0) {
    __write2("could not create raw socket. euid==0?\n");
    return 1;
  }

  while ((c = getopt (argc, argv, "i:")) != -1)
    switch (c) {
    case 'i':
      interface=if_nametoindex(optarg);
      if (!interface) {
	__write2("interface not found!\n");
	return 1;
      }
      break;
    default:
usage:
	      __write2("usage: arping [-i interface] ip\n");
	      return 0;
    }

  if (argc>optind)
    peer=argv[optind];
  else
    goto usage;

  memset(&sout,0,sizeof(sout));
  memset(&sin,0,sizeof(sout));
  sout.sll_family=AF_PACKET;
  sout.sll_ifindex=interface;
  if (bind(s,(struct sockaddr*)&sout,sizeof(sout))==-1) {
    __write2("bind failed!  Use -i eth1 to set interface.\n");
    return 1;
  }

  sl=sizeof(sin);
  if (getsockname(s,(struct sockaddr*)&sin,&sl)==-1) {
    __write2("getsockname failed?!?\n");
    return -1;
  }
  if (sin.sll_hatype!=1) {
    __write2("not Ethernet?!\n");
    return -1;
  }
  memcpy(arprequest+6,sin.sll_addr,6);
  memcpy(arprequest+22,sin.sll_addr,6);

  sout.sll_protocol=htons(ETH_P_ARP);

  {
    char ip[4];
    if (peer[scan_ip4(peer,ip)]) {
      __write2("bad target IP.\n");
      return 1;
    }
    memcpy(arprequest+38,ip,4);
  }

  __write1("ARPING ");
  __write1(peer);
  __write1("... (1 arp request per second)\n");
  missed=0;

  for (;;) {
    struct timeval deadline,now,diff;
    struct pollfd p;
    if (sendto(s,arprequest,42,0,(struct sockaddr*)&sout,sizeof(sout))==-1) {
      __write2("Can't happen error: sendto failed.  Sorry.\n");
      return 1;
    }
    gettimeofday(&deadline,0);
    ++deadline.tv_sec;
    p.fd=s;
    p.events=POLLIN|POLLERR;
    for (;;) {
      long timeout;
      int l;
      if (missed>3) {
	char buf[100];
	__write1("[");
	write(1,buf,fmt_ulong(buf,missed));
	__write1(" unanswered queries]\r");
      }
      gettimeofday(&now,0);
      if (now.tv_sec>deadline.tv_sec || (now.tv_sec==deadline.tv_sec && now.tv_usec>deadline.tv_usec))
	break;
      timeout=(deadline.tv_sec-now.tv_sec)*1000000 +
	      (deadline.tv_usec-now.tv_usec);
      if (poll(&p,1,timeout/1000)==0) {
	++missed;
	break;
      }
      sl=sizeof(sin);
      if ((l=recvfrom(s,ibuf,sizeof(ibuf),0,(struct sockaddr*)&sin,&sl))==-1) {
	__write2("Can't happen error: recvfrom failed.  Sorry.\n");
	return 1;
      }
      /* too short to be a valid ARP reply? */
      if (l<42) continue;
      /* is it an ARP packet? */
      if (ibuf[12]!=arprequest[12] || ibuf[13]!=arprequest[13]) continue;
      /* is it an ARP reply? */
      if (ibuf[20]!=0 || ibuf[21]!=2) continue;
      /* is it the IP address I asked for? */
      if (memcmp(ibuf+28,arprequest+38,4)) continue;
      if (missed>3) __write1("\n");
      missed=0;
      __write1("ARP reply from ");
      dumpethaddr(ibuf+32);
      __write1(" (");
      {
	char buf[100];
	write(1,buf,fmt_ip4(buf,ibuf+38));
      }
      __write1(")\n");
    }
  }
}
